;;;; dollar.scm
;;;; Kon Lovett, Mar '09
;
; Copyright (c) 2006, Felix L. Winkelmann
; All rights reserved.
;
; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
; conditions are met:
;
;   Redistributions of source code must retain the above copyright notice, this list of conditions and the following
;     disclaimer.
;   Redistributions in binary frm must reproduce the above copyright notice, this list of conditions and the following
;     disclaimer in the documentation and/or other materials provided with the distribution.
;   Neither the name of the author nor the names of its contributors may be used to endorse or promote
;     products derived from this software without specific prior written permission.
;
; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
; OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
; AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR
; CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
; CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
; SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
; OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
; POSSIBILITY OF SUCH DAMAGE.
;
; Send bugs, suggestions and ideas to:
;
; felix@call-with-current-continuation.org
;
; Felix L. Winkelmann
; Unter den Gleichen 1
; 37130 Gleichen
; Germany

;;; The dollar macro

(module dollar ($)

  (import scheme chicken foreign)
  (begin-for-syntax
   (require-extension srfi-4))

;;

(define-syntax $
  (er-macro-transformer
    (lambda (frm rnm cmp)
      (##sys#check-syntax '$ frm '(_ symbol . #(_ 0)))
      (let ((_quote (rnm 'quote))
            (_location (rnm 'location))
            (_void (rnm 'void))
            (_bool (rnm 'bool))
            (_char (rnm 'char))
            (_int (rnm 'int))
            (_double (rnm 'double))
            (_symbol (rnm 'symbol))
            (_scheme-object (rnm 'scheme-object))
            (_nonnull-c-string (rnm 'nonnull-c-string))
            (_nonnull-c-pointer (rnm 'nonnull-c-pointer))
            (_nonnull-u8vector (rnm 'nonnull-u8vector))
            (_nonnull-s8vector (rnm 'nonnull-s8vector))
            (_nonnull-u16vector (rnm 'nonnull-u16vector))
            (_nonnull-s16vector (rnm 'nonnull-s16vector))
            (_nonnull-u32vector (rnm 'nonnull-u32vector))
            (_nonnull-s32vector (rnm 'nonnull-s32vector))
            (_nonnull-f32vector (rnm 'nonnull-f32vector))
            (_nonnull-f64vector (rnm 'nonnull-f64vector))
            (_foreign-code (rnm 'foreign-code))
            (_foreign-value (rnm 'foreign-value))
            (_foreign-lambda* (rnm 'foreign-lambda*)))

      (define (unknown-type-error x) (syntax-error '$ "bad argument type - unsupported" x))

      (define (ensure-typed-atom val)
        (cond ((fixnum? val)      `(,_int ,val))
              ((number? val)      `(,_double ,val))
              ((string? val)      `(,_nonnull-c-string ,val))
              ((char? val)        `(,_char ,val))
              ((boolean? val)     `(,_bool ,val))
              ((null? val)        `(,_scheme-object ,val))
              ((eof-object? val)  `(,_scheme-object ,val))
              (else
               (unknown-type-error val)) ) )

      (define (pair-or-vector? x) (or (pair? x) (vector? x)))

      (define (ensure-typed-arg arg)
        (cond ((null? arg)
               (syntax-error '$ "bad argument type - missing type" arg) )
              ((atom? arg)
               (ensure-typed-atom arg))
              ((list? arg)
               (let ((typ (car arg)))
                 (cond ((cmp _quote typ)
                        (if (not (pair? (cdr arg))) arg
                            (let ((val (cadr arg)))
                              (cond ((symbol? val)      `(,_symbol ',val))
                                    ((u8vector? val)    `(,_nonnull-u8vector ',val))
                                    ((s8vector? val)    `(,_nonnull-s8vector ',val))
                                    ((u16vector? val)   `(,_nonnull-u16vector ',val))
                                    ((s16vector? val)   `(,_nonnull-s16vector ',val))
                                    ((u32vector? val)   `(,_nonnull-u32vector ',val))
                                    ((s32vector? val)   `(,_nonnull-s32vector ',val))
                                    ((f32vector? val)   `(,_nonnull-f32vector ',val))
                                    ((f64vector? val)   `(,_nonnull-f64vector ',val))
                                    ((pair-or-vector? val) `(,_scheme-object ',val))
                                    (else                (ensure-typed-atom val) ) ) ) ) )
                       ((cmp _location typ) `(,_nonnull-c-pointer ,arg))
                       ((not (pair? (cdr arg))) (ensure-typed-atom arg))
                       (else (list typ (cadr arg)) ) ) ) )
              (else
               (unknown-type-error arg) ) ) )

      (define (gentmp _) (rnm (gensym)))

      (define (symstr x) (symbol->string (strip-syntax x)))

      (let* ((func (cadr frm))
             (args (cddr frm))
             (rtype (cond ((and (pair? args) (symbol? (car args)))
                           (let ((rtype func))
                             (set! func (car args))
                             (set! args (cdr args))
                             rtype))
                          (else 'void)))
             (cargs (map ensure-typed-arg args))
             ; Gen tmp sym for each C argument
             (tmps (map gentmp cargs) ) )
        (cond ((not (null? cargs))
               `((,_foreign-lambda*
                  ,rtype
                  ,(map (lambda (carg tmp) (list (car carg) tmp)) cargs tmps)
                  ,(let ((fcall (conc func #\( (string-intersperse (map symstr tmps) ",") #\))))
                     (conc (if (cmp _void rtype) fcall (string-append "return(" fcall ")")) #\;)))
                 ,@(map cadr cargs)) )
              ((cmp _void rtype)
               `(,_foreign-code ,(conc func #\( #\) #\;) void) )
              (else
               `(,_foreign-value ,(conc func #\( #\)) ,rtype) ) ) ) ) ) ) )

) ;module dollar
